/*
 * TENCO
 * 
 * File encryption/decryption tool.
 * Copyright-free 2004-2014 by Masahiko Nakamura.
 * 
 * ver 1.1a/1.0a
 *          2014/08/14  64bit support
 * ver 1.1  2010/09/07  change iv
 * ver 1.0  2010/05/05  change header, random unused
 * ver 0.6  2007/11/11  Open source
 * ver 0.5  2005/10/02  randomize, change check-method
 * ver 0.4  2005/03/03  Linux
 * ver 0.3  2004/06/14  AES
 * ver 0.2  2004/04/04  Windows, Towfish
 * ver 0.1  2004/03/27  alpha version
 */

#if !defined(__unix__)
	#define	_CRT_SECURE_NO_DEPRECATE
//	#define	_USE_32BIT_TIME_T
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#if defined(__unix__)
	#include <unistd.h>
	#include <utime.h>
	#include <sys/time.h>
	#define	_stat		stat
	#define	_utimbuf	utimbuf
	#define	_utime		utime
	#define	_cputs(s)	fputs(s, stderr)
	#define	_getch()	getchar()
	#define	_putch(c)	fputc(c, stderr)
	#define	_MAX_PATH	260
#else
	#include <conio.h>
	#include <io.h>
	#include <sys/timeb.h>
	#include <sys/utime.h>
#endif

#include "rijndael-alg-fst.h"
#include "sha1.h"
#include "twofish.h"

#define	ENCRYPT	1
#define	DECRYPT	2
#define	SELFENC	3
#define	SELFDEC	4
#define	TCFID	"TCF"
#define	IDLEN	3
#define	TCFVER_TWOFISH	0x01
#define	TCFVER_AES		0x02
#define	TCFVER_SETIV	0x04
#define	TCFVER_RANDOM	0x08
#define	TCFVER_NEWHEAD	0x10
#define	TCFVER_NEWIV	0x20
#define	TCFVER_ALGOMASK	0x03
#define	MAXVER	(TCFVER_AES | TCFVER_SETIV | TCFVER_NEWHEAD)
#define	BLOCK	16
#define	TCFEXT	".tcf"
#define	PATHENV	"PATH"
#define	UNIXATTRFLG	0x40000000
#define	NOTCONFIRM	0x20000000	/* obsolete */
#define	TMPNAME	"_tenco_.tmp"

#if defined(__unix__)
	#define	SELFEXT	".tce"
	#define	ECHO_ON()	system("stty echo >/dev/null 2>&1")
	#define	ECHO_OFF()	system("stty -echo >/dev/null 2>&1")
	#define	PATHSEP	':'
	#define	FILESEP	'/'
	#define	FILESEPSTR	"/"
#else
	#define	SELFEXT	".exe"
	#define	ECHO_ON()
	#define	ECHO_OFF()
	#define	FILESEP	'\\'
#endif

struct tcfheader_t {
	char id[3];
	char ver;
	char size_new[4];
	char etc[4];
	char size_old[4];
	char time[4];
	char attr[4];
	char name[256];
	char sha1[20];
};

int mode = 0;
int algo = 0;
char tcfver = 0;

char outname[_MAX_PATH];
char *infile;
char *outfile;
FILE *in = 0;
FILE *out = 0;
FILE *self = 0;

Twofish_Byte key[32];
int key_len;
Twofish_key xkey;
Twofish_Byte p[BLOCK];
Twofish_Byte c[BLOCK];

int aeskeyLen;
int aeskeyNr;
u32 aeskeyRk[4*(MAXNR + 1)];

/* confirm */
void confirm(void)
{
	if (mode) {
		_cputs("hit any key....");
		ECHO_OFF();
		_getch();
		ECHO_ON();
		_cputs("\r               \r");
	}
	
	return;
}

/* input key string */
int inputkey(char *prompt, char *buff, int len)
{
	int i;
	int j;
	int n;
	int ch;
	
	memset(buff, 0, len);
	_cputs(prompt);
	n = strlen(prompt);
	ECHO_OFF();
	for (i = 0; i < len - 1; i++) {
		ch = _getch();
		if (ch == 0x03) {	/* ^C */
			memset(buff, 0, len);
			i = -1;
			break;
		}
		if (ch == 0x0d || ch == 0x0a) {	/* ^M */
			break;
		}
		buff[i] = (char)ch;
#if !defined(__unix__)
//		_putch('*');
//		n++;
#endif
	}
	ECHO_ON();
	_putch('\r');
	for (j = 0; j < n; j++) {
		_putch(' ');
	}
	_putch('\r');
	
	return i;
}

/* clean up */
void cleanup(void)
{
	if (in) {
		fclose(in);
		in = 0;
	}
	if (out) {
		fclose(out);
		out = 0;
	}
	if (self) {
		fclose(self);
		self = 0;
	}
	
	return;
}

/* command line error */
void cmderror(void)
{
	cleanup();
	puts("TENCO ver 1.0a");
	puts("usage1: tenco <e|ea|d|s|sa> <infile> [key] [outfile]");
	puts("usage2: self-decrypt [key] [outfile]");
	puts("        e  = encrypt (Twofish)");
	puts("        ea = encrypt (AES)");
	puts("        d  = decrypt");
	puts("        s  = make self-decrypt file (Twofish)");
	puts("        sa = make self-decrypt file (AES)");
	puts("        [key] = max 32 bytes string");
	confirm();
	exit(1);
}

/* system error */
void syserror(char *errmsg)
{
	perror(errmsg);
	cleanup();
	if (outfile && *outfile) {
		remove(outfile);
	}
	confirm();
	exit(2);
}

/* application error */
void apperror(char *errmsg)
{
	puts(errmsg);
	cleanup();
	if (outfile && *outfile) {
		remove(outfile);
	}
	confirm();
	exit(3);
}

/* twofish error */
void twofisherror(char *errmsg)
{
	puts(errmsg);
	cleanup();
	if (outfile && *outfile) {
		remove(outfile);
	}
	confirm();
	exit(4);
}

/* read input-file */
int readin(void *buff, int num)
{
	int readnum;
	
	readnum = fread(buff, sizeof(char), num, in);
	if (ferror(in)) {
		syserror("infile read error");
	}
	
	return readnum;
}

/* write output-file */
int writeout(void *buff, int num)
{
	int writenum;
	
	writenum = fwrite(buff, sizeof(char), num, out);
	if (ferror(out)) {
		syserror("outfile write error");
	}
	
	return writenum;
}

/* initialize */
void init(char *keystr)
{
	algo = tcfver & TCFVER_ALGOMASK;
	if (algo != TCFVER_TWOFISH && algo != TCFVER_AES) {
		apperror("infile version error.");
	}
	key_len = strlen(keystr);
	if (key_len > sizeof(key)) {
		key_len = sizeof(key);
	}
	memset(key, 0, sizeof(key));
	memcpy(key, keystr, key_len);
	if (algo == TCFVER_TWOFISH) {
		Twofish_initialise();
		memset(&xkey, 0, sizeof(xkey));
		Twofish_prepare_key(key, key_len, &xkey);
	}
	else {
		aeskeyLen = (key_len + 7) >> 3;
		if (aeskeyLen < 2) {
			aeskeyLen = 2;
		}
		aeskeyLen *= 64;
		memset(aeskeyRk, 0, sizeof(aeskeyRk));
		if (mode == ENCRYPT || mode == SELFENC) {
			aeskeyNr = rijndaelKeySetupEnc(aeskeyRk, key, aeskeyLen);
		}
		else {
			aeskeyNr = rijndaelKeySetupDec(aeskeyRk, key, aeskeyLen);
		}
	}
	
	return;
}

/* make iv */
void makeiv(void)
{
	int i;
	
	memset(p, 0, sizeof(p));
	memset(c, 0, sizeof(c));
	/* Fill in the array c with your own value. */
	/* It means the password only for your program. */
	if (tcfver & TCFVER_SETIV) {
		for (i = 0; i < sizeof(c); i++) {
			c[i] = (Twofish_Byte)(0x7f - i);
		}
	}
	if (tcfver & TCFVER_NEWIV) {
		/* secret by the author */
	}
	
	return;
}

/* encrypt */
void encrypt(unsigned char *buff)
{
	int i;
	
	for (i = 0; i < BLOCK; i++) {
		buff[i] ^= c[i];
	}
	if (algo == TCFVER_TWOFISH) {
		Twofish_encrypt(&xkey, buff, c);
	}
	else {
		rijndaelEncrypt(aeskeyRk, aeskeyNr, buff, c);
	}
	memcpy(buff, c, BLOCK);
	
	return;
}

/* decrypt */
void decrypt(unsigned char *buff)
{
	int i;
	
	if (algo == TCFVER_TWOFISH) {
		Twofish_decrypt(&xkey, buff, p);
	}
	else {
		rijndaelDecrypt(aeskeyRk, aeskeyNr, buff, p);
	}
	for (i = 0; i < BLOCK; i++) {
		p[i] ^= c[i];
	}
	memcpy(c, buff, BLOCK);
	memcpy(buff, p, BLOCK);
	
	return;
}

/* multi-byte(kanji) strchr */
char *kstrchr(char *str, int c)
{
#if defined(__unix__)
	return strchr(str, c);
#else 
	while (*str) {
		int t = (int)(unsigned char)(*str);
		if (t == c) {
			return str;
		}
		if ((t >= 0x81 && t <= 0x9F) || (t >= 0xE0 && t <= 0xFC)) {
			str++;
			if (!*str) {
				break;
			}
			str++;
		}
		else {
			str++;
		}
	}
	
	return 0;
#endif
}

/* random number */
int timerand(void)
{
	int randnum;
	
#if defined(__unix__)
	struct timeval tv;
	struct timezone tz;
	
	gettimeofday(&tv, &tz);
	randnum = tv.tv_sec * 1000 + tv.tv_usec / 1000;
#else 
	struct _timeb time;
	
	_ftime(&time);
	randnum = (int)time.time * 1000 + time.millitm;
#endif
	
	return randnum;
}

/* main routine */
int main(int argc, char *argv[])
{
	char *keystr;
	unsigned int outoffset;
	struct tcfheader_t header;
	unsigned char buff[BLOCK];
	unsigned char sha1[20];
	unsigned int len;
	unsigned int readnum;
	unsigned int writenum;
	char keybuff[32 + 1];
	char keyconf[32 + 1];
	char work[256];
	int i;
	
	/* clear vars */
	keystr = 0;
	outoffset = 0;
	memset(&header, 0, sizeof(header));
	
	/* selffile open */
#if defined(__unix__)
	self = fopen(argv[0], "rb");
	if (!self) {
		char selffile[_MAX_PATH];
		char *ppath;
		char *ptmp;
		
		ptmp = ppath = getenv(PATHENV);
		while (ptmp) {
			ptmp = kstrchr(ppath, PATHSEP);
			if (ptmp) {
				memset(selffile, 0, sizeof(selffile));
				memcpy(selffile, ppath, ptmp - ppath);
				strcat(selffile, FILESEPSTR);
				strcat(selffile, argv[0]);
				self = fopen(selffile, "rb");
				if (self) {
					break;
				}
				ptmp++;
				ppath = ptmp;
			}
			else {
				memset(selffile, 0, sizeof(selffile));
				strcpy(selffile, ppath);
				strcat(selffile, FILESEPSTR);
				strcat(selffile, argv[0]);
				self = fopen(selffile, "rb");
				if (self) {
					break;
				}
			}
		}
		if (!self) {
			syserror("selffile open error");
		}
	}
#else 
	if (!self) {
		char selffile[_MAX_PATH];
		
		strcpy(work, argv[0]);
		strcat(work, SELFEXT);
		_searchenv(work, PATHENV, selffile);
		if (!*selffile) {
			strcpy(work, argv[0]);
			_searchenv(work, PATHENV, selffile);
		}
		if (!*selffile) {
			apperror("selffile find error.");
		}
		self = fopen(selffile, "rb");
		if (!self) {
			syserror("selffile open error");
		}
	}
#endif
	
	/* self-decrypt check */
	memset(work, 0, IDLEN + 1);
	while ((i = fgetc(self)) != EOF) {
		if (memcmp(work, TCFID, IDLEN) == 0 && i > 0) {
			if (fseek(self, -(IDLEN + 1), SEEK_CUR)) {
				syserror("selffile seek error");
			}
			mode = SELFDEC;
			break;
		}
		work[IDLEN] = (char)i;
		memmove(work, &work[1], IDLEN);
	}
	
	/* command line check */
	if (mode == SELFDEC) {
		if (argc > 1) {
			keystr = argv[1];
		}
		if (argc > 2) {
			outfile = argv[2];
		}
	}
	if (!mode && argc > 2) {
		if (strcmp(argv[1], "e") == 0) {
			mode = ENCRYPT;
			tcfver = TCFVER_TWOFISH | TCFVER_SETIV | TCFVER_NEWHEAD;
		}
		else if (strcmp(argv[1], "ea") == 0) {
			mode = ENCRYPT;
			tcfver = TCFVER_AES | TCFVER_SETIV | TCFVER_NEWHEAD;
		}
		else if (strcmp(argv[1], "d") == 0) {
			mode = DECRYPT;
		}
		else if (strcmp(argv[1], "s") == 0) {
			mode = SELFENC;
			tcfver = TCFVER_TWOFISH | TCFVER_SETIV | TCFVER_NEWHEAD;
		}
		else if (strcmp(argv[1], "sa") == 0) {
			mode = SELFENC;
			tcfver = TCFVER_AES | TCFVER_SETIV | TCFVER_NEWHEAD;
		}
		infile = argv[2];
		for (i = 3; i < argc; i++) {
			if (strcmp(argv[i], "-nc") == 0) {
				/* obsolete */
			}
			else if (!keystr) {
				keystr = argv[i];
			}
			else if (!outfile) {
				outfile = argv[i];
			}
		}
	}
	if (!mode) {
		cmderror();
	}
	
	/* key check */
	if (!keystr) {
		memset(keybuff, 0, sizeof(keybuff));
		if (mode == ENCRYPT || mode == SELFENC) {
			for (;;) {
				if (inputkey("enter key:", keybuff, sizeof(keybuff)) == -1) {
					apperror("aborted.");
				}
				if (inputkey("retype key:", keyconf, sizeof(keyconf)) == -1) {
					apperror("aborted.");
				}
				if (strcmp(keybuff, keyconf) == 0) {
					break;
				}
				puts("keys mismatch.");
			}
		}
		else {
			if (inputkey("enter key:", keybuff, sizeof(keybuff)) == -1) {
				apperror("aborted.");
			}
		}
		keystr = keybuff;
	}
	
	if (mode == ENCRYPT || mode == SELFENC) {
		struct _stat instat;
		char *pname;
		char *ptmp;
		int tmp;
		
		/* startup */
		init(keystr);
		makeiv();
		sha1init();
		
		/* infile open */
		in = fopen(infile, "rb");
		if (!in) {
			syserror("infile open error");
		}
		
		/* infile info */
		memset(&instat, 0, sizeof(instat));
		if (_stat(infile, &instat) == -1) {
			syserror("infile info error");
		}
		
		/* infile name */
		ptmp = pname = infile;
		while (ptmp) {
			ptmp = kstrchr(pname, FILESEP);
			if (ptmp) {
				ptmp++;
				pname = ptmp;
			}
		}
		if (strlen(pname) >= sizeof(header.name)) {
			apperror("infile name error.");
		}
		
		/* outfile name */
		if (!outfile) {
			if (strlen(pname) + strlen(TCFEXT) >= sizeof(outname)) {
				apperror("infile name error.");
			}
			memset(outname, 0, sizeof(outname));
			strcpy(outname, pname);
			if (mode == ENCRYPT) {
				strcat(outname, TCFEXT);
			}
			else if (mode == SELFENC) {
				strcat(outname, SELFEXT);
			}
			outfile = outname;
		}
		
		/* outfile open */
		out = fopen(outfile, "wb");
		if (!out) {
			syserror("outfile open error");
		}
		
		/* write self-file */
		if (mode == SELFENC) {
			if (fseek(self, 0, SEEK_SET)) {
				syserror("selffile seek error");
			}
			while (!feof(self)) {
				readnum = fread(work, sizeof(char), sizeof(work), self);
				if (ferror(self)) {
					syserror("selffile read error");
				}
				writenum = fwrite(work, sizeof(char), readnum, out);
				if (ferror(out)) {
					syserror("outfile write error");
				}
				outoffset += writenum;
			}
		}
		
		/* write header */
		memcpy(header.id, TCFID, sizeof(header.id));
		header.ver = tcfver;
		if (tcfver & TCFVER_RANDOM) {
			tmp = timerand();
			memcpy(header.etc, &tmp, sizeof(header.etc));
		}
		tmp = instat.st_size;
		memcpy(header.size_new, &tmp, sizeof(header.size_new));
		tmp = (int)instat.st_mtime;
		memcpy(header.time, &tmp, sizeof(header.time));
		tmp = instat.st_mode;
#if defined(__unix__)
		tmp |= UNIXATTRFLG;
#endif
		memcpy(header.attr, &tmp, sizeof(header.attr));
		strcpy(header.name, pname);
		writeout(&header, sizeof(header));
		
		/* encrypt! */
		len = instat.st_size;
		while (len) {
			memset(buff, 0, sizeof(buff));
			readnum = readin(buff, sizeof(buff));
			if (readnum < sizeof(buff) && readnum < len) {
				apperror("infile size error.");
			}
			sha1data(buff, sizeof(buff));
			encrypt(buff);
			writeout(buff, sizeof(buff));
			len -= readnum;
		}
		
		/* rewrite header */
		sha1data((void *)header.etc, BLOCK);
		encrypt((void *)header.etc);
		for (i = 0; i < sizeof(header.name); i += BLOCK) {
			sha1data((void *)&header.name[i], BLOCK);
			encrypt((void *)&header.name[i]);
		}
		len = header.etc - header.id;
		if (fseek(out, outoffset + len, SEEK_SET)) {
			syserror("outfile seek error");
		}
		len = header.sha1 - header.etc;
		writeout(&header.etc, len);
		
		/* write sha1 */
		sha1hash(sha1);
		len = header.sha1 - header.id;
		if (fseek(out, outoffset + len, SEEK_SET)) {
			syserror("outfile seek error");
		}
		writeout(sha1, sizeof(sha1));
		
		/* set executable */
#if defined(__unix__)
		if (mode == SELFENC) {
			struct _stat outstat;
			
			memset(&outstat, 0, sizeof(outstat));
			if (fstat(fileno(out), &outstat) == -1) {
				syserror("outfile info error");
			}
			if (outstat.st_mode & S_IRUSR) {
				outstat.st_mode |= S_IXUSR;
			}
			if (outstat.st_mode & S_IRGRP) {
				outstat.st_mode |= S_IXGRP;
			}
			if (outstat.st_mode & S_IROTH) {
				outstat.st_mode |= S_IXOTH;
			}
			if (fchmod(fileno(out), outstat.st_mode) == -1) {
				syserror("outfile mode error");
			}
		}
#endif
		
		/* clean up */
		cleanup();
		
		/* finish */
		puts("encrypted.");
	}
	else if (mode == DECRYPT || mode == SELFDEC) {
		int filenew;
		
		if (mode == DECRYPT) {
			/* file open */
			in = fopen(infile, "rb");
			if (!in) {
				syserror("infile open error");
			}
			
			/* self-decrypt check */
			memset(work, 0, IDLEN + 1);
			while ((i = fgetc(in)) != EOF) {
				if (memcmp(work, TCFID, IDLEN) == 0 && i > 0) {
					if (fseek(in, -(IDLEN + 1), SEEK_CUR)) {
						syserror("infile seek error");
					}
					break;
				}
				work[IDLEN] = (char)i;
				memmove(work, &work[1], IDLEN);
			}
		}
		else if (mode == SELFDEC) {
			/* file change */
			in = self;
			self = 0;
		}
		
		/* read header 1/2 */
		len = header.size_new - header.id;
		readnum = readin(&header, len);
		if (readnum < len
			|| memcmp(header.id, TCFID, sizeof(header.id)) != 0) {
			apperror("infile header error.");
		}
		
		/* startup */
		if (header.ver > MAXVER) {
			apperror("infile version error.");
		}
		tcfver = header.ver;
		init(keystr);
		makeiv();
		sha1init();
		
		/* read header 2/2 */
		if (tcfver & TCFVER_NEWHEAD) {
			len = sizeof(header) - (header.size_new - header.id);
			readnum = readin(&header.size_new, len);
		}
		else {
			len = sizeof(header) - (header.etc - header.id);
			readnum = readin(&header.etc, len);
		}
		if (readnum < len) {
			apperror("infile header error.");
		}
		if (!(tcfver & TCFVER_NEWHEAD)) {
			decrypt((void *)header.etc);
			sha1data((void *)header.etc, BLOCK);
			for (i = 0; i < sizeof(header.name); i += BLOCK) {
				decrypt((void *)&header.name[i]);
				sha1data((void *)&header.name[i], BLOCK);
			}
		}
		
		/* outfile name */
		if (outfile) {
			filenew = 1;
		}
		else {
			filenew = 0;
			strcpy(outname, TMPNAME);
			outfile = outname;
		}
		
		/* outfile open */
		out = fopen(outfile, "wb");
		if (!out) {
			syserror("outfile open error");
		}
		
		/* decrypt! */
		if (tcfver & TCFVER_NEWHEAD) {
			len = *((unsigned int *)header.size_new);
		}
		else {
			len = *((unsigned int *)header.size_old);
		}
		while (len) {
			readnum = readin(buff, sizeof(buff));
			if (readnum < sizeof(buff)) {
				apperror("infile data error.");
			}
			decrypt(buff);
			sha1data(buff, sizeof(buff));
			if (len > sizeof(buff)) {
				writenum = sizeof(buff);
			}
			else {
				writenum = len;
			}
			writeout(buff, writenum);
			len -= writenum;
		}
		
		/* reread header */
		if (tcfver & TCFVER_NEWHEAD) {
			decrypt((void *)header.etc);
			sha1data((void *)header.etc, BLOCK);
			for (i = 0; i < sizeof(header.name); i += BLOCK) {
				decrypt((void *)&header.name[i]);
				sha1data((void *)&header.name[i], BLOCK);
			}
		}
		
		/* check sha1 */
		sha1hash(sha1);
		if (memcmp(sha1, header.sha1, sizeof(sha1) != 0)) {
			apperror("infile data error.");
		}
		
		/* clean up */
		cleanup();
		
		/* set file name and time */
		if (!filenew) {
			struct _utimbuf times;
			
			remove(header.name);
			if (rename(outfile, header.name) != 0) {
				syserror("outfile name error");
			}
			memset(&times, 0, sizeof(times));
			memcpy(&times.actime, header.time, sizeof(header.time));
			memcpy(&times.modtime, header.time, sizeof(header.time));
			if (_utime(header.name, &times) == -1) {
				syserror("outfile time error");
			}
		}
		
		/* finish */
		puts("decrypted.");
	}
	
	return 0;
}

